/*
 * Copyright 2017 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */
 
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <ssdef.h>
#include <descrip.h>
#include <lib$routines.h>

#include "dk_vajhoej_vms_call_VMS.h"

#define MAX_ITEMS 100

struct item32 {
                short buflen,code;
                long bufadr,retlenadr;
                };

/* create descriptor */
static struct dsc$descriptor_s create_desc(const char *buf)
{
   struct dsc$descriptor_s res;
   res.dsc$w_length = strlen(buf);
   res.dsc$b_dtype = DSC$K_DTYPE_T;
   res.dsc$b_class = DSC$K_CLASS_S;
   res.dsc$a_pointer = (char *)buf;
   return res;
}

/* cached method id's */
static jmethodID mid_TypeWrapper_getValue;
static jmethodID mid_Short_shortValue;
static jmethodID mid_Integer_intValue;
static jmethodID mid_Long_longValue;
static jmethodID mid_Float_floatValue;
static jmethodID mid_Double_doubleValue;
static jmethodID mid_TypeWrapper_setValue_short;
static jmethodID mid_TypeWrapper_setValue_int;
static jmethodID mid_TypeWrapper_setValue_long;
static jmethodID mid_TypeWrapper_setValue_float;
static jmethodID mid_TypeWrapper_setValue_double;
static jmethodID mid_Item32_getType;
static jmethodID mid_Item32_getBuflen;
static jmethodID mid_Item32_getCode;
static jmethodID mid_Item32_getBufadr;
static jmethodID mid_Item32_getRetlenadrorflag;

#ifdef DEBUG
static char clsnambuf[1024];

static char *get_class_name(JNIEnv *cntx, jobject obj) {
    jclass cls;
    jmethodID mid;
    jobject clsobj;
    jclass clscls;
    jmethodID clsmid;
    jstring str;
    const char *p;
    cls = (*cntx)->GetObjectClass(cntx, obj);
    mid = (*cntx)->GetMethodID(cntx, cls, "getClass", "()Ljava/lang/Class;");
    clsobj = (*cntx)->CallObjectMethod(cntx, obj, mid);
    clscls = (*cntx)->GetObjectClass(cntx, clsobj);
    clsmid = (*cntx)->GetMethodID(cntx, clscls, "getName", "()Ljava/lang/String;");
    str = (jstring)(*cntx)->CallObjectMethod(cntx, clsobj, clsmid);
    p = (*cntx)->GetStringUTFChars(cntx, str, NULL);
    strcpy(clsnambuf, p);
    (*cntx)->ReleaseStringUTFChars(cntx, str, p);
    return clsnambuf;
}
#endif

static int rbuflen;
static char *rbuf;
/* ROUTINE0 support (LBR$GET_HELP) */
static long routine0(struct dsc$descriptor_s *a1, long *a2, long *a3, long a4)
{
	int ix;
	ix = strlen(rbuf);
	if(ix + a1->dsc$w_length > rbuflen) return SS$_ABORT;
	memcpy(rbuf + ix, a1->dsc$a_pointer, a1->dsc$w_length);
	ix += a1->dsc$w_length;
	rbuf[ix] = '\0';
	return SS$_NORMAL;
}

/* get short/int/long/float/double/descriptor from Short/Integer/Long/Float/Double/byte[] */
static short get_w(JNIEnv *cntx, jobject sobj)
{
   return (*cntx)->CallShortMethod(cntx, sobj, mid_Short_shortValue);
}

static long get_lw(JNIEnv *cntx, jobject iobj)
{
   return (*cntx)->CallIntMethod(cntx, iobj, mid_Integer_intValue);
}

static long long get_qw(JNIEnv *cntx, jobject lobj)
{
   return (*cntx)->CallLongMethod(cntx, lobj, mid_Long_longValue);
}

static float get_sf(JNIEnv *cntx, jobject fobj)
{
   return (*cntx)->CallFloatMethod(cntx, fobj, mid_Float_floatValue);
}

static double get_tf(JNIEnv *cntx, jobject dobj)
{
   return (*cntx)->CallDoubleMethod(cntx, dobj, mid_Double_doubleValue);
}

static struct dsc$descriptor_s desc_get_ba(JNIEnv *cntx, jobject baobj)
{
   jbyte* buf;
   jsize len;
   struct dsc$descriptor_s res;
   buf = (*cntx)->GetByteArrayElements(cntx, baobj, NULL);
   len = (*cntx)->GetArrayLength(cntx, baobj);
   res.dsc$w_length = len;
   res.dsc$b_dtype = DSC$K_DTYPE_T;
   res.dsc$b_class = DSC$K_CLASS_S;
   res.dsc$a_pointer = (char *)buf;
   return res;
}

static char *get_ba(JNIEnv *cntx, jobject baobj)
{
   jbyte* buf;
   buf = (*cntx)->GetByteArrayElements(cntx, baobj, NULL);
   return (char *)buf;
}

/* get Object from TypeWrapper */
static jobject get_wrapped(JNIEnv *cntx, jobject obj)
{
   return (*cntx)->CallObjectMethod(cntx, obj, mid_TypeWrapper_getValue);
}

/* get short/int/long/float/double/byte[] from TypeWrapper */
static short get_wrapped_w(JNIEnv *cntx, jobject sobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap w from %s\n", get_class_name(cntx, sobj));
#endif
   return get_w(cntx, get_wrapped(cntx, sobj));
}

static long get_wrapped_lw(JNIEnv *cntx, jobject iobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap lw from %s\n", get_class_name(cntx, iobj));
#endif
   return get_lw(cntx, get_wrapped(cntx, iobj));
}

static long long get_wrapped_qw(JNIEnv *cntx, jobject lobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap qw from %s\n", get_class_name(cntx, lobj));
#endif
   return get_qw(cntx, get_wrapped(cntx, lobj));
}

static float get_wrapped_sf(JNIEnv *cntx, jobject fobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap sf from %s\n", get_class_name(cntx, fobj));
#endif
   return get_sf(cntx, get_wrapped(cntx, fobj));
}

static double get_wrapped_tf(JNIEnv *cntx, jobject dobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap tf from %s\n", get_class_name(cntx, dobj));
#endif
   return get_tf(cntx, get_wrapped(cntx, dobj));
}

static struct dsc$descriptor_s desc_get_wrapped_ba(JNIEnv *cntx, jobject baobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap ba from %s\n", get_class_name(cntx, baobj));
#endif
   return desc_get_ba(cntx, get_wrapped(cntx, baobj));
}

static char *get_wrapped_ba(JNIEnv *cntx, jobject baobj)
{
#ifdef DEBUG
   fprintf(stderr, "unwrap ba from %s\n", get_class_name(cntx, baobj));
#endif
   return get_ba(cntx, get_wrapped(cntx, baobj));
}

static int get_ba_length(JNIEnv *cntx, jobject baobj)
{
#ifdef DEBUG
   fprintf(stderr, "get ba length from %s\n", get_class_name(cntx, baobj));
#endif
   return (*cntx)->GetArrayLength(cntx, get_wrapped(cntx, baobj));
}

/* set short/int/long/float/double/byte[] in TypeWrapper */
static void set_wrapped_w(JNIEnv *cntx, jobject sobj, short v)
{
#ifdef DEBUG
   fprintf(stderr, "wrap w into %s\n", get_class_name(cntx, sobj));
#endif
   (*cntx)->CallObjectMethod(cntx, sobj, mid_TypeWrapper_setValue_short, v);
}

static void set_wrapped_lw(JNIEnv *cntx, jobject iobj, long v)
{
#ifdef DEBUG
   fprintf(stderr, "wrap lw into %s\n", get_class_name(cntx, iobj));
#endif
   (*cntx)->CallObjectMethod(cntx, iobj, mid_TypeWrapper_setValue_int, v);
}

static void set_wrapped_qw(JNIEnv *cntx, jobject lobj, long long v)
{
#ifdef DEBUG
   fprintf(stderr, "wrap qw into %s\n", get_class_name(cntx, lobj));
#endif
   (*cntx)->CallObjectMethod(cntx, lobj, mid_TypeWrapper_setValue_long, v);
}

static void set_wrapped_sf(JNIEnv *cntx, jobject fobj, float v)
{
#ifdef DEBUG
   fprintf(stderr, "wrap sf into %s\n", get_class_name(cntx, fobj));
#endif
   (*cntx)->CallObjectMethod(cntx, fobj, mid_TypeWrapper_setValue_float, v);
}

static void set_wrapped_tf(JNIEnv *cntx, jobject dobj, double v)
{
#ifdef DEBUG
   fprintf(stderr, "wrap tf into %s\n", get_class_name(cntx, dobj));
#endif
   (*cntx)->CallObjectMethod(cntx, dobj, mid_TypeWrapper_setValue_double, v);
}

static void desc_set_wrapped_ba(JNIEnv *cntx, jobject baobj, struct dsc$descriptor_s v, int wrtbck)
{
#ifdef DEBUG
   fprintf(stderr, "wrap ba into %s\n", get_class_name(cntx, baobj));
#endif
   (*cntx)->ReleaseByteArrayElements(cntx, baobj, (signed char *)v.dsc$a_pointer, wrtbck ? 0 : JNI_ABORT);
}

static void set_wrapped_ba(JNIEnv *cntx, jobject baobj, char *v, int wrtbck)
{
#ifdef DEBUG
   fprintf(stderr, "wrap ba into %s\n", get_class_name(cntx, baobj));
#endif
   (*cntx)->ReleaseByteArrayElements(cntx, baobj, (signed char *)v, wrtbck ? 0 : JNI_ABORT);
}

/* get type from Item32 */
static long get_item_type(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallIntMethod(cntx, lstobj, mid_Item32_getType);
}

static long get_item_buflen(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallIntMethod(cntx, lstobj, mid_Item32_getBuflen);
}

static long get_item_code(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallIntMethod(cntx, lstobj, mid_Item32_getCode);
}

static jobject get_item_bufadr(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallObjectMethod(cntx, lstobj, mid_Item32_getBufadr);
}

static jobject get_item_retlenadrorflag(JNIEnv *cntx, jobject lstobj)
{
     return (*cntx)->CallObjectMethod(cntx, lstobj, mid_Item32_getRetlenadrorflag);
}

/* load cache of method id's */
JNIEXPORT jboolean JNICALL Java_dk_vajhoej_vms_call_VMS_setup(JNIEnv *cntx, jclass me)
{
   jclass cls;
   /* TypeWrapper.getValue */
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$TypeWrapper");
   if(cls == 0) return 0;
   mid_TypeWrapper_getValue = (*cntx)->GetMethodID(cntx, cls, "getValue", "()Ljava/lang/Object;");
   if(mid_TypeWrapper_getValue == 0) return 0;
   /* Short.shortValue/Integer.intValue/Long.longValue */
   cls = (*cntx)->FindClass(cntx, "java/lang/Short");
   if(cls == 0) return 0;
   mid_Short_shortValue = (*cntx)->GetMethodID(cntx, cls, "shortValue", "()S");
   if(mid_Short_shortValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Integer");
   if(cls == 0) return 0;
   mid_Integer_intValue = (*cntx)->GetMethodID(cntx, cls, "intValue", "()I");
   if(mid_Integer_intValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Long");
   if(cls == 0) return 0;
   mid_Long_longValue = (*cntx)->GetMethodID(cntx, cls, "longValue", "()J");
   if(mid_Long_longValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Float");
   if(cls == 0) return 0;
   mid_Float_floatValue = (*cntx)->GetMethodID(cntx, cls, "floatValue", "()F");
   if(mid_Float_floatValue == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "java/lang/Double");
   if(cls == 0) return 0;
   mid_Double_doubleValue = (*cntx)->GetMethodID(cntx, cls, "doubleValue", "()D");
   if(mid_Double_doubleValue == 0) return 0;
   /* Word.setValue/LongWord.setValue/QuadWord.setValue */
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$Word");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_short = (*cntx)->GetMethodID(cntx, cls, "setValue", "(S)V");
   if(mid_TypeWrapper_setValue_short == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$LongWord");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_int = (*cntx)->GetMethodID(cntx, cls, "setValue", "(I)V");
   if(mid_TypeWrapper_setValue_int == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$QuadWord");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_long = (*cntx)->GetMethodID(cntx, cls, "setValue", "(J)V");
   if(mid_TypeWrapper_setValue_long == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$SFloat");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_float = (*cntx)->GetMethodID(cntx, cls, "setValue", "(F)V");
   if(mid_TypeWrapper_setValue_float == 0) return 0;
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$TFloat");
   if(cls == 0) return 0;
   mid_TypeWrapper_setValue_double = (*cntx)->GetMethodID(cntx, cls, "setValue", "(D)V");
   if(mid_TypeWrapper_setValue_double == 0) return 0;
   /* Item32.getXxxx */
   cls = (*cntx)->FindClass(cntx, "dk/vajhoej/vms/call/VMS$Item32");
   if(cls == 0) return 0;
   mid_Item32_getType = (*cntx)->GetMethodID(cntx, cls, "getType", "()I");
   if(mid_Item32_getType == 0) return 0;
   mid_Item32_getBuflen = (*cntx)->GetMethodID(cntx, cls, "getBuflen", "()I");
   if(mid_Item32_getBuflen == 0) return 0;
   mid_Item32_getCode = (*cntx)->GetMethodID(cntx, cls, "getCode", "()I");
   if(mid_Item32_getCode == 0) return 0;
   mid_Item32_getBufadr = (*cntx)->GetMethodID(cntx, cls, "getBufadr", "()Ljava/lang/Object;");
   if(mid_Item32_getBufadr == 0) return 0;
   mid_Item32_getRetlenadrorflag = (*cntx)->GetMethodID(cntx, cls, "getRetlenadrorflag", "()Ljava/lang/Object;");
   if(mid_Item32_getRetlenadrorflag == 0) return 0;
   return 1;
}

/* find entrypoint in shareable image */
JNIEXPORT jlong JNICALL Java_dk_vajhoej_vms_call_VMS_getFunction(JNIEnv *cntx, jclass me, jstring image, jstring function)
{
   int result;
   int f;
   const char *image2;
   const char *function2;
   struct dsc$descriptor_s imagedesc;
   struct dsc$descriptor_s functiondesc;
   image2 = (*cntx)->GetStringUTFChars(cntx, image, 0);
   imagedesc = create_desc(image2);
   function2 = (*cntx)->GetStringUTFChars(cntx, function, 0);
   functiondesc = create_desc(function2);
   result = lib$find_image_symbol(&imagedesc, &functiondesc, &f);
#ifdef DEBUG
   fprintf(stderr, "LIB$FIND_IMAGE_SYMBOL image=%s function=%s result=%d f=%ld\n", image2, function2, result, f);
#endif
   (*cntx)->ReleaseStringUTFChars(cntx, image, image2);
   (*cntx)->ReleaseStringUTFChars(cntx, function, function2);
   return f;
}

/* call */
JNIEXPORT jint JNICALL Java_dk_vajhoej_vms_call_VMS_callFunction(JNIEnv *cntx, jclass me, jlong function, jintArray argtyp, jintArray argacc, jobjectArray argval)
{
   int narg;
   int nitem;
   int typ;
   int *arglist;
   int *temp;
   int *temp2;
   struct item32 *tempitmlst;
   long long temp8;
   float tempf4;
   double tempf8;
   struct dsc$descriptor_s tempd;
   int i;
   int j;
   int *p;
   jint *real_argtyp;
   jint *real_argacc;
   jobject real_argval;
   jobject real_argval2;
   jint result;
   /* setup */
   narg = (*cntx)->GetArrayLength(cntx, argtyp);
   arglist = calloc(1 + narg, sizeof(int));
   temp = calloc(narg * 2, sizeof(int));
   temp2 = calloc(MAX_ITEMS * 2, sizeof(int));
   tempitmlst = calloc(MAX_ITEMS, sizeof(struct item32));
#ifdef DEBUG
   fprintf(stderr, "call function=%lld\n", function);
#endif
   arglist[0] = narg;
   real_argtyp = (*cntx)->GetIntArrayElements(cntx, argtyp, 0);
   real_argacc = (*cntx)->GetIntArrayElements(cntx, argacc, 0);
   /* process arguments before calling native function */
   for(i = 0; i < narg; i++) 
   {
      real_argval = (*cntx)->GetObjectArrayElement(cntx, argval, i);
#ifdef DEBUG
       fprintf(stderr, "  arg=%d typ=%d acc=%d cls=%s\n", i+1, real_argtyp[i], real_argacc[i], get_class_name(cntx, real_argval));
#endif
      switch(real_argtyp[i]) 
      {
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYVALUE:
            arglist[1 + i] = get_wrapped_lw(cntx,real_argval);
#ifdef DEBUG
            fprintf(stderr, "    lwbv=%d\n", arglist[1 + i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_WORD_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                temp[2*i] = get_wrapped_w(cntx,real_argval);
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    wbr=%d\n", temp[2*i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                temp[2*i] = get_wrapped_lw(cntx,real_argval);
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    lwbr=%d\n", temp[2*i]);
#endif
            break;                    
         case dk_vajhoej_vms_call_VMS_QUADWORD_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                temp8 = get_wrapped_qw(cntx,real_argval);
                memcpy(&temp[2*i], &temp8, sizeof(long long));
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    qwbr=%lld\n", temp8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_SFLOAT_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                tempf4 = get_wrapped_sf(cntx,real_argval);
                memcpy(&temp[2*i], &tempf4, sizeof(float));
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    sfbr=%f\n", tempf4);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_TFLOAT_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_WRITEONLY) {
                tempf8 = get_wrapped_tf(cntx,real_argval);
                memcpy(&temp[2*i], &tempf8, sizeof(double));
            }
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    tfbr=%f\n", tempf8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_CHARACTERSTRING_BYDESCRIPTOR:
            tempd = desc_get_wrapped_ba(cntx,real_argval);
            memcpy(&temp[2*i], &tempd, sizeof(struct dsc$descriptor_s));
            arglist[i + 1] = (int)&temp[2*i];
#ifdef DEBUG
            fprintf(stderr, "    csbd=%s\n", tempd.dsc$a_pointer);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_BLOCK_BYREFERENCE:
             arglist[i + 1] = (int)get_wrapped_ba(cntx,real_argval);;
#ifdef DEBUG
            int *p = (int *)arglist[i + 1];
            fprintf(stderr, "    bbr=%08X %08X\n", p[0], p[1]);
#endif
         	 break;
         case dk_vajhoej_vms_call_VMS_ITEMLIST32_BYREFERENCE:
             real_argval = get_wrapped(cntx, real_argval);
             nitem = (*cntx)->GetArrayLength(cntx, real_argval);
             /* process item list */
             for(j = 0; j < nitem; j++)
             {
                 real_argval2 = (*cntx)->GetObjectArrayElement(cntx, real_argval, j);
                 typ = get_item_type(cntx, real_argval2);
#ifdef DEBUG
                 fprintf(stderr, "  item=%d typ=%d cls=%s\n", j+1, typ, get_class_name(cntx, real_argval2));
#endif
            	 tempitmlst[j].buflen = get_item_buflen(cntx, real_argval2);
            	 tempitmlst[j].code = get_item_code(cntx, real_argval2);
            	 switch(typ) {
					 case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_IN:
						 temp2[2*j] = get_wrapped_lw(cntx, get_item_bufadr(cntx, real_argval2));
						 tempitmlst[j].bufadr = (long)&temp2[2*j];
						 tempitmlst[j].retlenadr = get_lw(cntx, get_item_retlenadrorflag(cntx, real_argval2));
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_OUT:
						 temp2[2*j] = get_wrapped_lw(cntx, get_item_bufadr(cntx, real_argval2));
						 tempitmlst[j].bufadr = (long)&temp2[2*j];
						 tempitmlst[j].retlenadr = 0;
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_IN:
 						 temp8 = get_wrapped_qw(cntx, get_item_bufadr(cntx, real_argval2));
 						 memcpy(&temp2[2*j], &temp8, sizeof(long long));
						 tempitmlst[j].bufadr = (long)&temp2[2*j];
						 tempitmlst[j].retlenadr = get_lw(cntx, get_item_retlenadrorflag(cntx, real_argval2));
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_OUT:
 						 temp8 = get_wrapped_qw(cntx, get_item_bufadr(cntx, real_argval2));
 						 memcpy(&temp2[2*j], &temp8, sizeof(long long));
						 tempitmlst[j].bufadr = (long)&temp2[2*j];
						 tempitmlst[j].retlenadr = 0;
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_IN:
				         tempd = desc_get_wrapped_ba(cntx,get_item_bufadr(cntx,real_argval2));
				         memcpy(&temp2[2*j], &tempd, sizeof(struct dsc$descriptor_s));
			             tempitmlst[j].bufadr = temp2[2*j+1];
		            	 tempitmlst[j].retlenadr = get_lw(cntx, get_item_retlenadrorflag(cntx, real_argval2));
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_OUT:
				         tempd = desc_get_wrapped_ba(cntx,get_item_bufadr(cntx,real_argval2));
				         memcpy(&temp2[2*j], &tempd, sizeof(struct dsc$descriptor_s));
			             tempitmlst[j].bufadr = temp2[2*j+1];
						 temp2[2*j] = get_wrapped_lw(cntx, get_item_retlenadrorflag(cntx, real_argval2));
						 tempitmlst[j].retlenadr = (long)&temp2[2*j];
						 break;
            	 }
        	 }
        	 tempitmlst[nitem].buflen = 0;
        	 tempitmlst[nitem].code = 0;
        	 tempitmlst[nitem].bufadr = 0;
        	 tempitmlst[nitem].retlenadr = 0;
        	 arglist[i + 1] = (int)&tempitmlst[0];
#ifdef DEBUG
             for(j = 0; j <= nitem; j++)
             {
                 fprintf(stderr, "  %04X%04X\n", tempitmlst[j].code, tempitmlst[j].buflen);
                 if(tempitmlst[j].bufadr > 0x100)
                 {
                	 p = (int *)tempitmlst[j].bufadr;
                     fprintf(stderr, "  %08X -> %08X%08X\n", tempitmlst[j].bufadr,  p[0], p[1]);
                 }
                 else
                 {
                     fprintf(stderr, "  %08X\n", tempitmlst[j].bufadr);
                 }
                 fprintf(stderr, "  %08X\n", tempitmlst[j].retlenadr);
             }
#endif
             break;
         case dk_vajhoej_vms_call_VMS_ROUTINE_0:
        	 rbuflen = get_ba_length(cntx,real_argval);
        	 rbuf = get_wrapped_ba(cntx,real_argval);
             arglist[i + 1] = (int)&routine0;
 #ifdef DEBUG
             fprintf(stderr, "    r=%d rbuflen=%d\n", arglist[i + 1], rbuflen);
 #endif
        	 break;
      }
   }
#ifdef DEBUG
   fprintf(stderr, "  before call\n");
#endif
   /* actual call of native function */
   result = lib$callg(arglist, function);
#ifdef DEBUG
   fprintf(stderr, "  after call sts=%d\n", result);
#endif
   /* process arguments after calling native function */
   for(i = 0; i < narg; i++) 
   {
      real_argval = (*cntx)->GetObjectArrayElement(cntx, argval, i);
#ifdef DEBUG
       fprintf(stderr, "  arg=%d typ=%d acc=%d cls=%s\n", i+1, real_argtyp[i], real_argacc[i], get_class_name(cntx, real_argval));
#endif
      switch(real_argtyp[i]) 
      {
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYVALUE:
            /* nothing */
            break;
         case dk_vajhoej_vms_call_VMS_WORD_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                set_wrapped_w(cntx,real_argval,temp[2*i]);
            }
#ifdef DEBUG
            fprintf(stderr, "    wbr=%d\n", temp[2*i]);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_LONGWORD_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                set_wrapped_lw(cntx,real_argval,temp[2*i]);
            }
#ifdef DEBUG
            fprintf(stderr, "    lwbr=%d\n", temp[2*i]);
#endif
            break;                    
         case dk_vajhoej_vms_call_VMS_QUADWORD_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                memcpy(&temp8,&temp[2*i],sizeof(long long));
                set_wrapped_lw(cntx,real_argval,temp8);
            }
#ifdef DEBUG
            fprintf(stderr, "    qwbr=%lld\n", temp8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_SFLOAT_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                memcpy(&tempf4,&temp[2*i],sizeof(float));
                set_wrapped_sf(cntx,real_argval,tempf4);
            }
#ifdef DEBUG
            fprintf(stderr, "    sfbr=%f\n", tempf4);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_TFLOAT_BYREFERENCE:
            if(real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY) {
                memcpy(&tempf8,&temp[2*i],sizeof(double));
                set_wrapped_tf(cntx,real_argval,tempf8);
            }
#ifdef DEBUG
            fprintf(stderr, "    tfbr=%f\n", tempf8);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_CHARACTERSTRING_BYDESCRIPTOR:
            memcpy(&tempd,&temp[2*i],sizeof(long long));
            desc_set_wrapped_ba(cntx,real_argval,tempd,real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY);
#ifdef DEBUG
            fprintf(stderr, "    csbd=%s\n", tempd.dsc$a_pointer);
#endif
            break;
         case dk_vajhoej_vms_call_VMS_BLOCK_BYREFERENCE:
         	 set_wrapped_ba(cntx,real_argval,(char *)arglist[i + 1],real_argacc[i] != dk_vajhoej_vms_call_VMS_READONLY);
#ifdef DEBUG
            int *p = (int *)arglist[i + 1];
            fprintf(stderr, "    bbr=%08X %08X\n", p[0], p[1]);
#endif
         	 break;
         case dk_vajhoej_vms_call_VMS_ITEMLIST32_BYREFERENCE:
             real_argval = get_wrapped(cntx, real_argval);
             nitem = (*cntx)->GetArrayLength(cntx, real_argval);
             /* process item list */
             for(j = 0; j < nitem; j++)
             {
                 real_argval2 = (*cntx)->GetObjectArrayElement(cntx, real_argval, j);
                 typ = get_item_type(cntx, real_argval2);
#ifdef DEBUG
                 fprintf(stderr, "  item=%d typ=%d cls=%s\n", j+1, typ, get_class_name(cntx, real_argval2));
#endif
            	 switch(typ) {
					 case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_IN:
						 /* nothing */
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_LONGWORD_OUT:
						 set_wrapped_lw(cntx,get_item_bufadr(cntx, real_argval2),temp2[2*j]);
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_IN:
						 /* nothing */
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_QUADWORD_OUT:
  		                 memcpy(&temp8,&temp2[2*j],sizeof(long long));
						 set_wrapped_qw(cntx,get_item_bufadr(cntx, real_argval2),temp8);
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_IN:
						 memcpy(&tempd,&temp2[2*j],sizeof(long long));
						 desc_set_wrapped_ba(cntx,real_argval2,tempd,0);
						 break;
					 case dk_vajhoej_vms_call_VMS_ITEM_CHARACTERSTRING_OUT:
						 memcpy(&tempd,&temp2[2*j],sizeof(long long));
						 desc_set_wrapped_ba(cntx,real_argval2,tempd,1);
						 break;
            	 }
        	 }
#ifdef DEBUG
             for(j = 0; j <= nitem; j++)
             {
                 fprintf(stderr, "  %04X%04X\n", tempitmlst[j].code, tempitmlst[j].buflen);
                 if(tempitmlst[j].bufadr > 0x100)
                 {
                	 p = (int *)tempitmlst[j].bufadr;
                     fprintf(stderr, "  %08X -> %08X%08X\n", tempitmlst[j].bufadr,  p[0], p[1]);
                 }
                 else
                 {
                     fprintf(stderr, "  %08X\n", tempitmlst[j].bufadr);
                 }
                 fprintf(stderr, "  %08X\n", tempitmlst[j].retlenadr);
             }
#endif
             break;
         case dk_vajhoej_vms_call_VMS_ROUTINE_0:
        	 set_wrapped_ba(cntx,real_argval,rbuf,1);
#ifdef DEBUG
            fprintf(stderr, "    r rbuf=%s\n", rbuf);
#endif
        	 break;
      }
   }
   /* cleanup */
   (*cntx)->ReleaseIntArrayElements(cntx, argtyp, real_argtyp, 0);
   (*cntx)->ReleaseIntArrayElements(cntx, argacc, real_argacc, 0);
   free(arglist);
   free(temp);
   free(temp2);
   free(tempitmlst);
   return result;
}

